/***************************************************************************
% *   Copyright (C) 2009 by Kamaledin Ghiasi-Shirazi                        *
% *   ghiathi@gmail.com                                                     *
% *                                                                         *
% *   This code is developed over an implementation of Canny edge detector  *
% *   by Mike Heath (heath@csee.usf.edu).					    *
% *   									    *
% *   This program is free software; you can redistribute it and/or modify  *
% *   it under the terms of the GNU General Public License as published by  *
% *   the Free Software Foundation; either version 2 of the License, or     *
% *   (at your option) any later version.                                   *
% *                                                                         *
% *   This program is distributed in the hope that it will be useful,       *
% *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
% *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
% *   GNU General Public License for more details.                          *
% *                                                                         *
% *   You should have received a copy of the GNU General Public License     *
% *   along with this program; if not, write to the                         *
% *   Free Software Foundation, Inc.,                                       *
% *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
% ***************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define VERBOSE 1
#define BOOSTBLURFACTOR 90.0


int read_pgm_image(char *infilename, unsigned char **image, int *rows, int *cols);
int write_pgm_image(char *outfilename, unsigned char *image, int rows, int cols, char *comment, int maxval);
void omnidirectional_edge_detector(unsigned char *image, int rows, int cols, float* filter1D, int window_size, float tlow, float thigh, unsigned char **edge, char *fname);
void complex_filter(unsigned char *image, int rows, int cols,  float* filter, int window_size, short int *filteredim_real, short int *filteredim_imag);
void generalize_1D_filter_to_2D(float* filter1D, int window_size, float* filter2D_real, float *filter2D_imag);

void make_DoB_Filter1D(float* filter1D, int window_size);
void make_Demigny_Filter1D(float* filter1D, int window_size);

void magnitude_x_y(short int *filter2D_real, short int *filter2D_imag, int rows, int cols,  short int **magnitude);
void apply_hysteresis(short int *mag, unsigned char *nms, int rows, int cols, float tlow, float thigh, unsigned char *edge);
void radian_direction(short int *filter2D_real, short int *filter2D_imag, int rows,  int cols, float **dir_radians, int xdirtag, int ydirtag);
double angle_radians(double x, double y);


main(int argc, char *argv[])
{
   char *infilename = NULL;  /* Name of the input image */
   char *dirfilename = NULL; /* Name of the output direction image */
   char outfilename[1024];    /* Name of the output "edge" image */
   char composedfname[1024];  /* Name of the output "direction" image */
   unsigned char *image;     /* The input image */
   unsigned char *edge;      /* The output edge image */
   int rows, cols;           /* The dimensions of the image. */
   char filter[128]; 	     /* Name of 1D filter. */
   float filter1D[128];
   int	window_size;
   float tlow,               /* Fraction of the high threshold in hysteresis. */
	 thigh;              /* High hysteresis threshold control. The actual
			        threshold is the (100 * thigh) percentage point
			        in the histogram of the magnitude of the
			        gradient image that passes non-maximal
			        suppression. */

   /****************************************************************************
   * Get the command line arguments.
   ****************************************************************************/
   if(argc < 5){
      fprintf(stderr,"\n<USAGE> %s image window_size tlow thigh filterName\n",argv[0]);
      fprintf(stderr,"\n      image:      An image to process. Must be in PGM format.\n");
      fprintf(stderr,"      filter:  1D-filter name (Demigny,DoB,...)\n");
      fprintf(stderr,"      window_size:  Window Size\n");
      fprintf(stderr,"      tlow:       Fraction (0.0-1.0) of the high edge strength threshold.\n");
      fprintf(stderr,"      thigh:      Fraction (0.0-1.0) of the distribution of non-zero edge\n");
      fprintf(stderr,"                   strengths for ");
      fprintf(stderr,"hysteresis. The fraction is used to compute\n");
      fprintf(stderr,"                  the high edge strength threshold.\n\n");
      exit(1);
   }

   infilename = argv[1];
   strcpy (filter, argv[2]);
   window_size = atoi(argv[3]);
   tlow = atof(argv[4]);
   thigh = atof(argv[5]);

   if(VERBOSE) 
   {
     printf("Input parameters.\n");
     printf("infilename = %s.\n", infilename);
     printf("filter = %s.\n", filter);
     printf("window_size = %d.\n", window_size);
     printf("tlow = %f.\n", tlow);
     printf("thigh = %f.\n", thigh);
   }

   /****************************************************************************
   * Read in the image. This read function allocates memory for the image.
   ****************************************************************************/
   if(VERBOSE) printf("Reading the image %s.\n", infilename);
   if(read_pgm_image(infilename, &image, &rows, &cols) == 0){
      fprintf(stderr, "Error reading the input image, %s.\n", infilename);
      exit(1);
   }


   /****************************************************************************
   * Generalizing the 1D filter to 2D
   * Note that the negative part of the 1D filter is not given.
   ****************************************************************************/

   if ( strcmp(filter,"Demigny") == 0)
	{
   	make_Demigny_Filter1D(filter1D, window_size);
	}
   else if (strcmp(filter,"DoB") == 0)
	{
   	make_DoB_Filter1D(filter1D, window_size);
	}
   else
	{
        fprintf(stderr, "Error in filter name (Not DoG nor Demigny).\n");
        exit(1);
	}


   /****************************************************************************
   * Perform the edge detection. All of the work takes place here.
   ****************************************************************************/
   if(VERBOSE) printf("Starting Ghiasi edge detection.\n");
   omnidirectional_edge_detector(image, rows, cols, filter1D, window_size, tlow, thigh, &edge, dirfilename);

   /****************************************************************************
   * Write out the edge image to a file.
   ****************************************************************************/
   /*sprintf(outfilename, "%s_%s_s_%3.2f_l_%3.2f_h_%3.2f.pgm", infilename,filter,
      param, tlow, thigh);
  */
   sprintf(outfilename, "%s_%s_w=%d_l=%3.2f_h=%3.2f.pgm", infilename, filter, window_size, tlow, thigh);
   if(VERBOSE) printf("Writing the edge iname in the file %s.\n", outfilename);
   if(write_pgm_image(outfilename, edge, rows, cols, "", 255) == 0){
      fprintf(stderr, "Error writing the edge image, %s.\n", outfilename);
      exit(1);
   }
}




/*******************************************************************************
* PROCEDURE: omnidirectional
* PURPOSE: To perform omnidirectional edge detection.
* NAME: Sayed Kamaledin Ghiasi Shirazi
* DATE: 16/2/2007
*******************************************************************************/
void omnidirectional_edge_detector(unsigned char *image, int rows, int cols, float* filter1D, int window_size, float tlow, float thigh, unsigned char **edge, char *fname)
{
   FILE *fpdir=NULL;          /* File to write the gradient image to.     */
   unsigned char *nms;        /* Points that are local maximal magnitude. */
   short int *filteredim_real, 
             *filteredim_imag, 
             *magnitude;      /* The magnitude of the gadient image.      */
   int r, c, pos;
   float *dir_radians=NULL;   /* Gradient direction image.                */


   /****************************************************************************
   * Allocate Filtered Image
   ****************************************************************************/
   if((filteredim_real = (short int *) calloc(rows*cols,
         sizeof(short int))) == NULL){
      fprintf(stderr, "Error allocating the smoothed image.\n");
      exit(1);
   }

   if((filteredim_imag = (short int *) calloc(rows*cols,
         sizeof(short int))) == NULL){
      fprintf(stderr, "Error allocating the smoothed image.\n");
      exit(1);
   }



   /****************************************************************************
   * Perform complexvalued filtering
   ****************************************************************************/
   if(VERBOSE) printf("Complex-valued filtering of the image.\n");
   complex_filter(image, rows, cols, filter1D, window_size, filteredim_real, filteredim_imag);

   /****************************************************************************
   * This option to write out the direction of the edge gradient was added
   * to make the information available for computing an edge quality figure
   * of merit.
   ****************************************************************************/
   if(fname != NULL){
      /*************************************************************************
      * Compute the direction up the gradient, in radians that are
      * specified counteclockwise from the positive x-axis.
      *************************************************************************/
      radian_direction(filteredim_real, filteredim_imag, rows, cols, &dir_radians, -1, -1);

      /*************************************************************************
      * Write the gradient direction image out to a file.
      *************************************************************************/
      if((fpdir = fopen(fname, "wb")) == NULL){
         fprintf(stderr, "Error opening the file %s for writing.\n", fname);
         exit(1);
      }
      fwrite(dir_radians, sizeof(float), rows*cols, fpdir);
      fclose(fpdir);
      free(dir_radians);
   }

   /****************************************************************************
   * Compute the magnitude of the gradient.
   ****************************************************************************/
   if(VERBOSE) printf("Computing the magnitude of the gradient.\n");
   magnitude_x_y(filteredim_real, filteredim_imag, rows, cols, &magnitude);

   /****************************************************************************
   * Perform non-maximal suppression.
   ****************************************************************************/
   if(VERBOSE) printf("Doing the non-maximal suppression.\n");
   if((nms = (unsigned char *) calloc(rows*cols,sizeof(unsigned char)))==NULL){
      fprintf(stderr, "Error allocating the nms image.\n");
      exit(1);
   }

   non_max_supp(magnitude, filteredim_real, filteredim_imag, rows, cols, nms);

   /****************************************************************************
   * Use hysteresis to mark the edge pixels.
   ****************************************************************************/
   if(VERBOSE) printf("Doing hysteresis thresholding.\n");
   if((*edge=(unsigned char *)calloc(rows*cols,sizeof(unsigned char))) ==NULL){
      fprintf(stderr, "Error allocating the edge image.\n");
      exit(1);
   }

   apply_hysteresis(magnitude, nms, rows, cols, tlow, thigh, *edge);

   /****************************************************************************
   * Free all of the memory that we allocated except for the edge image that
   * is still being used to store out result.
   ****************************************************************************/
   free(filteredim_real);
   free(filteredim_imag);
   free(magnitude);
   free(nms);
}


/*******************************************************************************
* PROCEDURE: complex_filter
* PURPOSE: Filter the image using a complex-valued mask.
* NAME: Sayed kamaledin Ghiasi Shirazi
* DATE: 16/2/2007
*******************************************************************************/
void complex_filter(unsigned char *image, int rows, int cols, float* filter1D, int window_size, short int *filteredim_real, short int *filteredim_imag)
{
   int r, c, rr, cc,     /* Counter variables. */
      center;            /* Half of the window_size. */
   float *filter2D_real, /* Real part of complex filter. */
	 *filter2D_imag, /* Imaginary part of complex filter. */
         dot_real,       /* Dot product summing variable. */
	 dot_imag;


   center = window_size / 2;

   /****************************************************************************
   * Allocate memory for real and imaginary filters
   ****************************************************************************/

   if(VERBOSE) printf("      The 2D Filter has %d elements.\n", window_size*window_size);
   if((filter2D_real = (float *) calloc(window_size*window_size, sizeof(float))) == NULL){
      fprintf(stderr, "Error callocing the 2D filter array.\n");
      exit(1);
   }
   if((filter2D_imag = (float *) calloc(window_size*window_size, sizeof(float))) == NULL){
      fprintf(stderr, "Error callocing the 2D filter array.\n");
      exit(1);
   }

   generalize_1D_filter_to_2D(filter1D, window_size, filter2D_real, filter2D_imag);


   /****************************************************************************
   * Perform filtering
   ****************************************************************************/
   if(VERBOSE) printf("   Filtering the image.\n");

   for(r=0;r<rows;r++){   
      for(c=0;c<cols;c++){
         dot_real = 0.0;
	 dot_imag = 0.0;
         for(cc=(-center);cc<=center;cc++){
            for(rr=(-center);rr<=center;rr++){
               if(((r+rr) >= 0) && ((r+rr) < rows)){
		 if(((c+cc) >= 0) && ((c+cc) < cols)){
                  dot_real += (float)image[(r+rr)*cols+(c+cc)] * filter2D_real[(center+cc)*window_size + (center+rr)];
                  dot_imag += (float)image[(r+rr)*cols+(c+cc)] * filter2D_imag[(center+cc)*window_size + (center+rr)];
		 }
               }
            }
         }
	 filteredim_real[r*cols+c] = (short int)(dot_real*BOOSTBLURFACTOR + 0.5);
	 filteredim_imag[r*cols+c] = (short int)(dot_imag*BOOSTBLURFACTOR + 0.5);
      }
   }

   free(filter2D_real);
   free(filter2D_imag);
}



/*******************************************************************************
* PROCEDURE: generalize_1D_filter_to_2D
* PURPOSE: Create a two dimensional complex extention of filter.
* NAME: Sayed kamaledin Ghiasi Shirazi
* DATE: 17/2/2007
*******************************************************************************/

void generalize_1D_filter_to_2D(float* filter1D, int window_size, float* filter2D_real, float *filter2D_imag)
{
   int i, j, center, dd; 	/*discrete d*/
   float x, y, f_real, f_imag, d, sum = 0.0, f_r, alpha;
   center = window_size / 2;

   for(i=0;i<window_size;i++){
     for(j=0;j<window_size;j++){
      x = (float)(i - center);
      y = (float)(j - center);
      d = (sqrt (x*x+y*y));
      dd = (int)(d);
      alpha = d - dd;
      //alpha = 0;
      if (d != 0)
      {
	f_r = (1-alpha) * filter1D[dd] + alpha * filter1D[dd+1];
      	f_real = f_r * x / d;
     	f_imag = f_r * y / d;
      }
      else
      {
	f_real = 0;
	f_imag = 0;
      }
      filter2D_real[i * window_size + j] = f_real;
      filter2D_imag[i * window_size + j] = f_imag;
      sum += fabs(f_real);
     }
   }

   //printf ("\n");
   for(i=0;i<window_size;i++){
     for(j=0;j<window_size;j++){
	filter2D_real[i * window_size + j] /= sum;
      	//printf ("\t(%f)",filter2D_real[i * window_size + j]);
     }
   //printf ("\n");
   }

   //printf ("\n");
   for(i=0;i<window_size;i++){
     for(j=0;j<window_size;j++){
	filter2D_imag[i * window_size + j] /= sum;
      	//printf ("\t(%f)",filter2D_imag[i * window_size + j]);
     }
   //printf ("\n");
   }

   if(VERBOSE){
      printf("The real filter coefficients are:\n");
	for(i=0;i<window_size;i++){
     	   for(j=0;j<window_size;j++){
              printf("filter2D_real[%d] = %f\n", i, filter2D_real[i]);
	   }
	}

      printf("\nThe imaginary filter coefficients are:\n");
	for(i=0;i<window_size;i++){
     	   for(j=0;j<window_size;j++){
              printf("filter2D_imag[%d] = %f\n", i, filter2D_imag[i]);
	   }
	}
   }
}



/*******************************************************************************
* PROCEDURE: make_Demigny_Filter1D
* PURPOSE: Create a one dimensional Demigny Filter.
* NAME: Sayed kamaledin Ghiasi Shirazi
* DATE: 19/2/2007
*******************************************************************************/

void make_Demigny_Filter1D(float* filter1D, int window_size)
{
   int i;
   float x, fx, sum=0.0;
   int W = window_size / 2;
 
   // Since the corners of the window have a distance eual to sqrt(2)*W, we zero out more coefficients.
   memset (filter1D,0, (window_size+1) * sizeof(float));

   switch (W){
   case 1:
        filter1D[0] = 0;	filter1D[1] = 1;	
	break;
   case 2:
	filter1D[0] = 0;	filter1D[1] = 1;	filter1D[2] = 0.6262;
	filter1D[3] = 0;
	break;
   case 3:
	filter1D[0] = 0;	filter1D[1] = 1;	filter1D[2] = 0.7673;	filter1D[3] = 0.5;	filter1D[4] = 0;
	break;
   case 4:
	filter1D[0] = 0; 	 filter1D[1] = 1; 	  filter1D[2] = 0.822;	filter1D[3] = 0.6625; filter1D[4] = 0.4319; filter1D[5] = 0;
	break;
   case 5:
	filter1D[0] = 0;	filter1D[1] = 1;	filter1D[2] = 0.8494;
	filter1D[3] = 0.7333;  filter1D[4] = 0.6;      filter1D[5] = 0.3868;
	filter1D[6] = 0;
	break;
   case 6:
	filter1D[0] = 0;	filter1D[1] = 1;	filter1D[2] = 0.5584;
	filter1D[3] = 0.8686;     filter1D[4] = 0.7744;	filter1D[5] = 0.6825;
	filter1D[6] = 0.3561;	filter1D[7] = 0;
	break;
   default:
        fprintf(stderr, "W in Demigny filter > 6.\n");
   };

   if(VERBOSE){
      printf("The filter coefficients are:\n");
      for(i=0;i<=(window_size / 2) + 1;i++)
         printf("filter1D[%d] = %f\n", i, filter1D[i]);
   }

}


void make_DoB_Filter1D(float* filter1D, int window_size)
{
   int i;
   int W = window_size / 2;

   // Since the corners of the window have a distance eual to sqrt(2)*W, we zero out more coefficients.
   memset (filter1D,0, (window_size+1) * sizeof(float));
   
   if(VERBOSE) printf("      The 1D Filter has %d elements.\n", window_size);
   if((filter1D = (float *) calloc(window_size, sizeof(float))) == NULL){
      fprintf(stderr, "Error callocing the DoG filter array.\n");
      exit(1);
   }
   
   memset (filter1D,0, (W+1) * sizeof(float));

   for (i=0;i<W;i++)
	filter1D[i]=1;
}



/*******************************************************************************
* Procedure: radian_direction
* Purpose: To compute a direction of the gradient image from component dx and
* dy images. Because not all derriviatives are computed in the same way, this
* code allows for dx or dy to have been calculated in different ways.
*
* FOR X:  xdirtag = -1  for  [-1 0  1]
*         xdirtag =  1  for  [ 1 0 -1]
*
* FOR Y:  ydirtag = -1  for  [-1 0  1]'
*         ydirtag =  1  for  [ 1 0 -1]'
*
* The resulting angle is in radians measured counterclockwise from the
* xdirection. The angle points "up the gradient".
*******************************************************************************/
void radian_direction(short int *delta_x, short int *delta_y, int rows,
    int cols, float **dir_radians, int xdirtag, int ydirtag)
{
   int r, c, pos;
   float *dirim=NULL;
   double dx, dy;

   /****************************************************************************
   * Allocate an image to store the direction of the gradient.
   ****************************************************************************/
   if((dirim = (float *) calloc(rows*cols, sizeof(float))) == NULL){
      fprintf(stderr, "Error allocating the gradient direction image.\n");
      exit(1);
   }
   *dir_radians = dirim;

   for(r=0,pos=0;r<rows;r++){
      for(c=0;c<cols;c++,pos++){
         dx = (double)delta_x[pos];
         dy = (double)delta_y[pos];

         if(xdirtag == 1) dx = -dx;
         if(ydirtag == -1) dy = -dy;

         dirim[pos] = (float)angle_radians(dx, dy);
      }
   }
}

/*******************************************************************************
* FUNCTION: angle_radians
* PURPOSE: This procedure computes the angle of a vector with components x and
* y. It returns this angle in radians with the answer being in the range
* 0 <= angle <2*PI.
*******************************************************************************/
double angle_radians(double x, double y)
{
   double xu, yu, ang;

   xu = fabs(x);
   yu = fabs(y);

   if((xu == 0) && (yu == 0)) return(0);

   ang = atan(yu/xu);

   if(x >= 0){
      if(y >= 0) return(ang);
      else return(2*M_PI - ang);
   }
   else{
      if(y >= 0) return(M_PI - ang);
      else return(M_PI + ang);
   }
}

/*******************************************************************************
* PROCEDURE: magnitude_x_y
* PURPOSE: Compute the magnitude of the gradient. This is the square root of
* the sum of the squared derivative values.
* NAME: Mike Heath
* DATE: 2/15/96
*******************************************************************************/
void magnitude_x_y(short int *delta_x, short int *delta_y, int rows, int cols,
        short int **magnitude)
{
   int r, c, pos, sq1, sq2;

   /****************************************************************************
   * Allocate an image to store the magnitude of the gradient.
   ****************************************************************************/
   if((*magnitude = (short *) calloc(rows*cols, sizeof(short))) == NULL){
      fprintf(stderr, "Error allocating the magnitude image.\n");
      exit(1);
   }

   for(r=0,pos=0;r<rows;r++){
      for(c=0;c<cols;c++,pos++){
         sq1 = (int)delta_x[pos] * (int)delta_x[pos];
         sq2 = (int)delta_y[pos] * (int)delta_y[pos];
         (*magnitude)[pos] = (short)(0.5 + sqrt((float)sq1 + (float)sq2));
      }
   }

}

